// ----------------------------------------------------------------------------
// VertexFormat.h
//
//    This file defines the tools needed to specify and work with 
//    API independent vertex format. The vertex format is defined
//    as a typelist. Each type of this list is defining an attribute 
//    (ex: the second texture coordinate as 2 floats is the 
//     mvf_texcoord1_2f type - or MVF_TEXCOORD1<float,2>).
//    Each attribute is associated to a base attribute. 
//    (ex: mvf_texcoord1_2f and mvf_texcoord1_3d are both associated
//     with MVF_BASE_TEXCOORD1 because they both correspond to a definition
//    of the second texture coordinate).
//    Given an attribute T, the base attribute is T::base_attribute
//
//    A format can easily be described by using the FormatDescriptor
//    class templated by the vertex format. It allows to know which
//    base attributes are present, the size of each attribute, the 
//    size of a vertex with this format, the offset of an attribute
//    in a table of bytes representing a vertex.
//
//
//                                                    (c) Sylvain Lefebvre 2003
// ----------------------------------------------------------------------------

#ifndef __VERTEXFORMAT__
#define __VERTEXFORMAT__

// -------------------------------------------------
// namespace GPUmesh { // no namespace under VC7 2002 ... pb with templates
// -------------------------------------------------
// Vertex attributes
// -------------------------------------------------

#include "VertexFormat_base.h"

// -------------------------------------------------
// Sizeof vertex attributes and vertex format
// -------------------------------------------------

template <class TList> class MVF_sizeof;

template <> class MVF_sizeof< Loki::NullType > {public: enum {value = 0}; };

template <class TList> class MVF_sizeof
{ 
private:

  ASSERT_TYPELIST(TList);

  typedef typename TList::Head Head;
  typedef typename TList::Tail Tail;

public:

  enum {value = Head::size_of + MVF_sizeof<Tail>::value};
};

// -------------------------------------------------
// Check whether an attribute corresponds to an item
// -------------------------------------------------

template <class T,class Item> class MVF_is_item
{
private:
  template <class I> struct In
  {
    enum {value = 0};
  };
  template <> struct In<Item>
  {
    enum {value = 1};
  };
public:
  enum {value = In<T::base_attribute>::value};
};

// -------------------------------------------------
// Vertex attribute offset
// -------------------------------------------------

template <class TList,class T> class MVF_offset
{ 
private:
  ASSERT_TYPELIST(TList);

  typedef typename TList::Tail Tail;

  template<class TList1> struct In
  {
  private:
    ASSERT_TYPELIST(TList1);

    typedef typename TList1::Tail Tail;
    typedef typename TList1::Head Head;

    enum { temp = (MVF_offset<Tail,T>::value) };
  public:
    enum { value = ((temp == -1) ? -1 : (Head::size_of + temp))};
  };
  template<> struct In< Loki::Typelist<T,Tail> >
  {
    enum { value = 0 };
  };
  template<> struct In< Loki::NullType >
  {
    enum { value = -1 };
  };

public:
  enum {value = In<TList>::value}; 
};

// -------------------------------------------------
// Vertex attribute offset from item
// -------------------------------------------------

template <class TList,class Item> class MVF_offset_item
{ 
private:
  ASSERT_TYPELIST(TList);

  template<class TList1> struct In
  {
  private:
    ASSERT_TYPELIST(TList1);

    typedef typename TList1::Tail Tail;
    typedef typename TList1::Head Head;

    enum { stop = MVF_is_item<Head,Item>::value     };
    enum { next = MVF_offset_item<Tail,Item>::value };
    enum { temp = (stop == 0) ? ((next == -1) ? -1 : (Head::size_of + next)) : 0 };
  public:
    enum { value = temp};
  };

  template<> struct In< Loki::NullType >
  {
    enum { value = -1 };
  };

public:
  enum {value = In<TList>::value}; 
};
// -------------------------------------------------
template <class TList, class Item> struct MVF_index_item
{

private:
  ASSERT_TYPELIST(TList);

  template<class TList1> struct In
  {
  private:
    ASSERT_TYPELIST(TList1);

    typedef typename TList1::Tail Tail;
    typedef typename TList1::Head Head;
    
    enum { stop = MVF_is_item<Head,Item>::value    };
    enum { next = MVF_index_item<Tail, Item>::value};
    enum { temp = (stop == 0) ? ((next == -1) ? -1 : (1+next)) : 0 };
  public:
    enum { value = temp };
  };

  template<> struct In<Loki::NullType>
  {
    enum { value = -1 };
  };

public:
  enum { value = In<TList>::value };
};

// -------------------------------------------------
// FormatDescriptor
// -------------------------------------------------

#include "FormatDescriptor.h"

// -------------------------------------------------
// CompatibleFormat
//   test if MVF1 is compatible with MVF2
// -------------------------------------------------

template <class MVF1,class MVF2> class CompatibleFormat
{ 
private:
  ASSERT_TYPELIST(MVF1);
  ASSERT_TYPELIST(MVF2);

  typedef typename MVF1::Head Head;
  typedef typename MVF1::Tail Tail;

  template <class _MVF1> struct In
  {
  private:
    typedef typename _MVF1::Tail Tail;
    typedef typename _MVF1::Head Head;
    enum { temp = (CompatibleFormat<Tail,MVF2>::value) };
  public:
    enum { value = temp * ((MVF_index_item<MVF2,Head::base_attribute>::value > -1)? 1 : 0)};
  };

  template <> struct In< Loki::NullType >
  {
    enum { value = 1 };
  };

public:
  enum {value = In<MVF1>::value}; 
};

// -------------------------------------------------
//  define a macro helping to check format compatibility

#define CHECK_COMPATIBLE_FORMAT(mvf1,mvf2) { \
  const int __d=CompatibleFormat<mvf1,mvf2>::value; \
  STATIC_CHECK(__d,vertex_format_not_compatible); }

// -------------------------------------------------
// } // namespace GPUmesh
// -------------------------------------------------
#endif
// -------------------------------------------------
