// ----------------------------------------------------------------------------
// VertexFormat.h
//
//    This file defines the tools needed to specify and work with 
//    API independent vertex format. The vertex format is defined
//    as a typelist. Each type of this list is defining an attribute 
//    (ex: the second texture coordinate as 2 floats is the 
//     mvf_texcoord1_2f type - or MVF_TEXCOORD1<float,2>).
//    Each attribute is associated to a base attribute. 
//    (ex: mvf_texcoord1_2f and mvf_texcoord1_3d are both associated
//     with MVF_BASE_TEXCOORD1 because they both correspond to a definition
//    of the second texture coordinate).
//    Given an attribute T, the base attribute is T::base_attribute
//
//    A format can easily be described by using the FormatDescriptor
//    class templated by the vertex format. It allows to know which
//    base attributes are present, the size of each attribute, the 
//    size of a vertex with this format, the offset of an attribute
//    in a table of bytes representing a vertex.
//
//
//                                                    (c) Sylvain Lefebvre 2003
// ----------------------------------------------------------------------------

#ifndef __VERTEXFORMAT__
#define __VERTEXFORMAT__

#include <Typelist.h>

// -------------------------------------------------
namespace GPUmesh {
// -------------------------------------------------
// Vertex attributes
// -------------------------------------------------

#include "VertexFormat_base.h"

// -------------------------------------------------
// Sizeof vertex attributes and vertex format
// -------------------------------------------------

template <class TList> class MVF_sizeof;

template <> class MVF_sizeof< Loki::NullType > {public: enum {value = 0}; };

template <class Head,class Tail> class MVF_sizeof< Loki::Typelist<Head,Tail> >
{ 
private:

public:

  enum {value = Head::size_of + MVF_sizeof<Tail>::value};
};

// -------------------------------------------------
// Check whether an attribute corresponds to an item
// -------------------------------------------------

template <class T,class Item> class MVF_is_item
{
public:
  enum {value = 0};
};

template <class T> class MVF_is_item<T,typename T::base_attribute>
{
public:
  enum {value = 1};
};

// -------------------------------------------------
// Vertex attribute offset
// -------------------------------------------------

template <class TList,class T> class MVF_offset;

template <class T> class MVF_offset<Loki::NullType,T>
{
public:
 enum { value = -1 }; 
};

template <class Tail,class T> class MVF_offset<Loki::Typelist<T,Tail>,T>
{
public:
 enum { value = 0 };   
};

template <class Head,class Tail,class T> class MVF_offset<Loki::Typelist<Head,Tail>,T>
{
public:
  enum { temp  = (MVF_offset<Tail,T>::value) };
  enum { value = ((temp == -1) ? -1 : (Head::size_of + temp))};
};

// -------------------------------------------------
// Vertex attribute offset from item
// -------------------------------------------------

template <class TList,class Item> class MVF_offset_item;

template <class Item> class MVF_offset_item<Loki::NullType,Item>
{
public:
  enum {value = -1};
};

template <class Head,class Tail,class Item> class MVF_offset_item<Loki::Typelist<Head,Tail>,Item>
{
private:
  enum { stop = MVF_is_item<Head,Item>::value     };
  enum { next = MVF_offset_item<Tail,Item>::value };
public:
  enum { value = (stop == 0) ? ((next == -1) ? -1 : (Head::size_of + next)) : 0 };
};

// -------------------------------------------------
// Retrieve index of an item
// -------------------------------------------------

template <class TList,class Item> class MVF_index_item;

template <class Item> class MVF_index_item<Loki::NullType,Item>
{
public:
  enum {value = -1};
};

template <class Head,class Tail,class Item> class MVF_index_item<Loki::Typelist<Head,Tail>,Item>
{
private:
    enum { stop = MVF_is_item<Head,Item>::value    };
    enum { next = MVF_index_item<Tail, Item>::value};
public:
    enum { value = (stop == 0) ? ((next == -1) ? -1 : (1+next)) : 0 };
};

// -------------------------------------------------
// FormatDescriptor
// -------------------------------------------------

#include "FormatDescriptor.h"

// -------------------------------------------------
// CompatibleFormat
//   test if MVF1 is compatible with MVF2
// -------------------------------------------------

template <class MVF1,class MVF2> class CompatibleFormat;

template <class MVF2> class CompatibleFormat<Loki::NullType,MVF2>
{
public:
  enum {value = 1};
};

template <class Head,class Tail,class MVF2> class CompatibleFormat<Loki::Typelist<Head,Tail>,MVF2>
{
  private:
    enum { temp = (CompatibleFormat<Tail,MVF2>::value) };
  public:
    enum { value = temp * ((MVF_index_item<MVF2,typename Head::base_attribute>::value > -1)? 1 : 0)};
};

// -------------------------------------------------
//  define a macro helping to check format compatibility

#define CHECK_COMPATIBLE_FORMAT(mvf1,mvf2) { \
  const int __d=CompatibleFormat<mvf1,mvf2>::value; \
  STATIC_CHECK(__d,vertex_format_not_compatible); }

// -------------------------------------------------
} // namespace GPUmesh
// -------------------------------------------------
#endif
// -------------------------------------------------
