// ----------------------------------------------------------------------------
// VertexStorage.h
//
//   A VertexStorage policy is used to specify how verticies should be stored 
//   by a RenderMesh class. There is basicly two ways of storing verticies and 
//   their attributes:
//
//   - interlaced (see VertexStorageInterlaced)
//     example:
//      P0_xyz N0_xyz T0_uv P1_xyz N1_xyz T1_uv ... Pn_xyz Nn_xyz Tn_uv
//   - separated  (see VertexStorageSeparateAttributes)
//     example:
//      P0_xyz P1_xyz ... Pn_xyz N0_xyz N1_xyz ... Nn_xyz T0_uv T1_uv ... Tn_uv
//
//   An additional policy, VertexCopy, is used to control vertex copy inside 
//   the storage policy. It allows to support plateform specific storage (like 
//   converting colors from RGBA to BGRA with Direct3D).
//
//
//                                                    (c) Sylvain Lefebvre 2003
// ----------------------------------------------------------------------------
#ifndef __VERTEXSTORAGE__
#define __VERTEXSTORAGE__
// -------------------------------------------------
#include "VertexFormat.h"
// -------------------------------------------------
#include <vector>
// -------------------------------------------------
namespace GPUmesh {
// -------------------------------------------------

template <class VertexFormat,typename vertex_data> class VertexCopyDefaultPolicy
{
public:
  static void copy(const vertex_data& src,vertex_data& dst) { dst=src; }
};

// -------------------------------------------------
// Store each attributes in separate vectors
// -------------------------------------------------

template <class VertexFormat,
          typename vertex_data,
          template <class,class> class VertexCopyPolicy=VertexCopyDefaultPolicy>
class VertexStorageSeparateAttributes
{
public:

  typedef FormatDescriptor<VertexFormat> vertex_format_desc;

  enum {nb_attributes=Loki::TL::Length<VertexFormat>::value}; 

  std::vector<unsigned char> m_Attributes[nb_attributes];
  int                        m_iVertices;

  VertexStorageSeparateAttributes() : m_iVertices(0) 
  {}

  void push_vertex(const vertex_data& v)
  {
    static vertex_data tmp;

    VertexCopyPolicy<VertexFormat,vertex_data>::copy(v,tmp);
    if (vertex_format_desc::has_normal)
    {
      for (int i=0;i<vertex_format_desc::type_normal::size_of;i++)
        m_Attributes[vertex_format_desc::idx_normal].push_back(tmp.data[vertex_format_desc::offset_normal+i]);
    }
    if (vertex_format_desc::has_color0)
    {
      for (int i=0;i<vertex_format_desc::type_color0::size_of;i++)
        m_Attributes[vertex_format_desc::idx_color0].push_back(tmp.data[vertex_format_desc::offset_color0+i]);
    }
    if (vertex_format_desc::has_color1)
    {
      for (int i=0;i<vertex_format_desc::type_color1::size_of;i++)
        m_Attributes[vertex_format_desc::idx_color1].push_back(tmp.data[vertex_format_desc::offset_color1+i]);
    }
    if (vertex_format_desc::has_texcoord0)
    {
      for (int i=0;i<vertex_format_desc::type_texcoord0::size_of;i++)
        m_Attributes[vertex_format_desc::idx_texcoord0].push_back(tmp.data[vertex_format_desc::offset_texcoord0+i]);
    }
    if (vertex_format_desc::has_texcoord1)
    {
      for (int i=0;i<vertex_format_desc::type_texcoord1::size_of;i++)
        m_Attributes[vertex_format_desc::idx_texcoord1].push_back(tmp.data[vertex_format_desc::offset_texcoord1+i]);
    }
    if (vertex_format_desc::has_vertex)
    {
      for (int i=0;i<vertex_format_desc::type_vertex::size_of;i++)
        m_Attributes[vertex_format_desc::idx_vertex].push_back(tmp.data[vertex_format_desc::offset_vertex+i]);
    }    
    m_iVertices++;
  }

  void clear()
  {
    for (int i=0;i<nb_attributes;i++)
      m_Attributes[i].clear();
    m_iVertices=0;
  }

  int nbVertices() const {return (m_iVertices);}
};

// -------------------------------------------------
// Store each vertex independently - attributes are interlaced
// -------------------------------------------------

template <class VertexFormat,
          typename vertex_data,
          template <class,class> class VertexCopyPolicy=VertexCopyDefaultPolicy>
class VertexStorageInterlaced
{
public:

  std::vector<vertex_data> m_Vertices;

  void push_vertex(const vertex_data& v)
  {
    static vertex_data tmp;

    VertexCopyPolicy<VertexFormat,vertex_data>::copy(v,tmp);
    m_Vertices.push_back(tmp);
  }

  void clear()
  {
    m_Vertices.clear();
  }

  int nbVertices() const {return ((int)m_Vertices.size());}

};
// -------------------------------------------------
} // namespace GPUmesh
// -------------------------------------------------
#endif
// -------------------------------------------------
